#!/usr/bin/python3
import tkinter as tk
from threading import Lock
import ovf_client_common as ovf
import os

# ----------------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------------
# Audio settings for different machines

audio_settings = {
    #'hostname': [pa_source, pa_sink, pa_source_volume, pa_sink_volume],

    'nepomuk': [
    'alsa_input.usb-0b0e_Jabra_Link_370_3050752DE03A015200-00.mono-fallback',
    'alsa_output.usb-0b0e_Jabra_Link_370_3050752DE03A015200-00.analog-stereo',
    100,
    70
    ],

    'poldi': [
    'alsa_input.pci-0000_00_14.2.analog-stereo',
    'alsa_output.pci-0000_00_14.2.analog-stereo',
    30,
    70
    ],
}

# Pulseaudio default settings
#
# Source (Microphone)
# Sink (Speaker)
pa_source = None # Unfortunately, there are no sinks or sources named 'default'.
pa_sink = None   # So we use the guess feature of pa_rtp.py. However , it
                 # works only in rare cases.
pa_source_volume = 30
pa_sink_volume = 77

# Read settings for hostname
hostname = os.uname()[1]
if hostname in audio_settings:
    print(f"Loading audio settings for '{hostname}'.")
    pa_source, pa_sink, pa_source_volume, pa_sink_volume = \
        audio_settings[hostname]


# ----------------------------------------------------------------------------
# GUI settings
# ----------------------------------------------------------------------------

# Text for GUI
#
text_quit = "aus"
text_connect = "verbunden"
text_rx = "RX"
text_tx = "TX"
text_listen = "mithören"
text_intern = "PTT intern"
text_mute_listen = "stumm/laut"
text_call_intern = "Ruf intern"
text_vol_plus = "Vol +"
text_vol_minus = "Vol -"
text_ptt = "PTT"

# Colours for GUI
#
c_blue = "#6060ff"
c_a_blue = "#c0c0ff"
c_red = "#ff0000"
c_a_red = "#ff8080"
c_green = "#00ff00"
c_a_green = "#20ff20"
c_yellow = "#ffe000"
c_a_yellow = "#fff0c0"

# ----------------------------------------------------------------------------
# class LED
# ----------------------------------------------------------------------------

class LED(tk.Button):
    """LED -- used as indicator"""

    def __init__(self, master=None, color=None, text=None):
        tk.Button.__init__(self, master,
                        relief = "ridge",
                        state='disabled',
                        text=text)
        # save colors
        self.bg_off = self["bg"]
        self.bg_on = color
        self.fg_off = self["disabledforeground"]
        self.fg_on = self["fg"]
        # start as off
        self.set(False)

    def set(self, state):
        """state = True/False for on/off"""
        if state:
            self.configure(disabledforeground=self.fg_on, bg=self.bg_on)
        else: 
            self.configure(disabledforeground=self.fg_off, bg=self.bg_off)

# ----------------------------------------------------------------------------
# class LatchButton
# ----------------------------------------------------------------------------

class LatchButton(tk.Button):
    """Button switch: press -> on, press -> off"""
    def __init__(self, master=None, 
            color=None, 
            activecolor=None, 
            text=None, 
            cmd_on=None, 
            cmd_off=None):
        tk.Button.__init__(self, master, 
                            command=self.push,
                            text=text)
        self.cmd_on = cmd_on
        self.cmd_off = cmd_off
        self.state = False
        # save colors
        self.bg_off = self["bg"]
        self.bga_off = self["activebackground"]
        self.bg_on = color
        self.bga_on = activecolor

    def push(self):
        """This method may be called externally to imitate putton press."""
        self.state = not self.state
        if self.state:
            self.configure(bg=self.bg_on, activebackground=self.bga_on)
            self.cmd_on()
        else:
            self.configure(bg=self.bg_off, activebackground=self.bga_off)
            self.cmd_off()

# ----------------------------------------------------------------------------
# Main program
# ----------------------------------------------------------------------------

vol_default = pa_sink_volume

class GUI:

    def __init__(self):
        # mechanic of gui
        self.state_ptt = False
        self.display_update_lock = Lock()
        self.display_update_req = False
        self.display_update = None

        # OVF client
        self.ovf = ovf.OVF_Client(self.ui_update_callback,
                             # Pulseaudio
                             pa_source, pa_sink,
                             pa_source_volume, pa_sink_volume)

        # Main Window
        self.root = tk.Tk()
        self.root.wm_title("OVF Client")

        # Frame containing widgets
        self.frame = tk.Frame(self.root)
        self.frame.pack()

        # Create widgets
        # Buttons
        self.B_quit = tk.Button(self.frame, 
                    text=text_quit,
                    fg='red',
                    command=self.root.destroy)
        self.B_mute_listen = tk.Button(self.frame, 
                    text=text_mute_listen,
                    bg=c_yellow,
                    activebackground=c_a_yellow,
                    command=self.ovf.ui_ev_press_listen)
        self.B_call_intern = tk.Button(self.frame, 
                    text=text_call_intern,
                    bg=c_blue,
                    activebackground=c_a_blue,
                    command=self.ovf.ui_ev_press_int)
        self.B_vol_plus = tk.Button(self.frame, 
                    text=text_vol_plus,
                    command=self.push_vol_plus)
        self.B_vol_minus = tk.Button(self.frame, 
                    text=text_vol_minus,
                    command=self.push_vol_minus)
        self.B_ptt = LatchButton(self.frame, 
                    color=c_red,
                    activecolor=c_a_red,
                    cmd_on=self.ovf.ui_ev_PTT_press,
                    cmd_off=self.ovf.ui_ev_PTT_release,
                    text=text_ptt)

        # Labels
        self.L_vol = tk.Label(self.frame,
                    justify="center",
                    text=self._vol_to_text(vol_default),
                    relief="sunken",
                    width=3)
        # LEDs
        self.L_rx = LED(self.frame,
                    color=c_green,
                    text=text_rx)
        self.L_tx = LED(self.frame,
                    color=c_red,
                    text=text_tx)
        self.L_listen = LED(self.frame,
                    color=c_yellow,
                    text=text_listen)
        self.L_intern = LED(self.frame,
                    color=c_blue,
                    text=text_intern)
        self.L_connect = LED(self.frame,
                    color=c_green,
                    text=text_connect)

        # Layout
        pady = 5
        padx = 5
        # Buttons
        self.B_quit.grid(row=0, column=0, 
                sticky='nsew', padx=padx, pady=pady)
        self.B_mute_listen.grid(row=2, column=1, 
                sticky='nsew', padx=padx, pady=pady)
        self.B_call_intern.grid(row=2, column=2, 
                sticky='nsew', padx=padx, pady=pady)
        self.B_vol_plus.grid(row=0, column=3, 
                sticky='nsew', padx=padx, pady=pady)
        self.B_vol_minus.grid(row=1, column=3, 
                sticky='nsew', padx=padx, pady=pady)
        self.B_ptt.grid(row=3, column=0, columnspan=5, 
                sticky='nsew', padx=padx, pady=pady)
        # Labels
        self.L_vol.grid(row=0, column=4, rowspan=2, 
                sticky='nsew', padx=padx, pady=pady)
        # LEDs
        self.L_rx.grid(row=0, column=1, 
                sticky='nsew', padx=padx, pady=pady)
        self.L_tx.grid(row=0, column=2, 
                sticky='nsew', padx=padx, pady=pady)
        self.L_listen.grid(row=1, column=1, 
                sticky='nsew', padx=padx, pady=pady)
        self.L_intern.grid(row=1, column=2, 
                sticky='nsew', padx=padx, pady=pady)
        self.L_connect.grid(row=1, column=0, 
                sticky='nsew', padx=padx, pady=pady)

        # save default colours
        self.def_bg_color = self.B_ptt["bg"]
        self.def_bga_color = self.B_ptt["activebackground"]

    def run(self):
        self.ovf.start()
        self.root.bind("<Key>", self.key_press)
        self.root.after(500, self.tk_callback_handler)
        self.root.mainloop()

    def done(self):
        # call done() for all childeren implementing it
        self.ovf.done()

    # --- volume -------------------------------------------------------------

    def push_vol_plus(self):
        self._push_vol(1)

    def push_vol_minus(self):
        self._push_vol(-1)

    def _push_vol(self, direction):
        self.ovf.ui_change_volume(direction * int(100/9))

    def _vol_to_text(self, vol):
        "Convert vol from [0..100] to nuber in [0..9]"
        return int(round(vol/100*9))

    # --- display update -----------------------------------------------------

    def ui_update_callback(self, RX, TX, Listen, Int, Con, Vol):
        """Callback for OVF_Client"""
        with self.display_update_lock:
            self.display_update = [RX, TX, Listen, Int, Con, Vol]
            self.display_update_req = True

    def tk_callback_handler(self):
        """Called periodically in TK's  mainloop().
           Performs request from ui_update_callback."""

        with self.display_update_lock:
            if self.display_update_req:
                self.display_update_req = False
                # update LEDs
                self.L_rx.set(self.display_update[0])
                self.L_tx.set(self.display_update[1])
                self.L_listen.set(self.display_update[2])
                self.L_intern.set(self.display_update[3])
                self.L_connect.set(self.display_update[4])

                # set volume
                Vol = self.display_update[5]
                self.L_vol.configure(text=self._vol_to_text(Vol))

        # restart callback_handler
        self.root.after(200, self.tk_callback_handler)


    # --- keyboard -----------------------------------------------------------

    def key_press(self, e):
        k = e.char
        if k == " " or k == "p":
            self.B_ptt.push()
        elif k == "+":
            self.push_vol_plus()
        elif k == "-":
            self.push_vol_minus()
        elif k== "l" or k == "s":
            self.ovf.ui_ev_press_listen()
        elif k== "i":
            self.ovf.ui_ev_press_int()
        elif k== "q":
            self.root.destroy()

my_gui = GUI()
my_gui.run()
my_gui.done()

# vim: set et:
