#!/usr/bin/python3
from threading import Thread, Timer, RLock, Event
from sys import argv, exit
import importlib
import socket
import time
import datetime
import physical_ptt
import thread_set
import pa_rtp
import th_audio_fft

verbose = False

# ----------------------------------------------------------------------------
# Config
# ----------------------------------------------------------------------------

# --- PTT setup --------------------------------------------------------------
#
# ptt_mode = 'rs232'    PTT is enabled via RTS line.
#                       ptt_port is serial port
#                       ('/dev/ttyUSB1' for FT-991A)
#
# ptt_mode = 'gpio'     PTT is enabled via GPIO pin (for Raspi)
#                       ptt_port is pin name
#                       ('GPIO26' for OVF-server hardware)
#
# ptt_delay             Delay in s to prolong the time the PTT is pushed on a
#                       release request. This is to avoid to cut-off the last
#                       word in a transmission due to digital audio delay.
ptt_mode = 'gpio'
ptt_port = 'GPIO24' # '/dev/ttyUSB1'
ptt_hold_on = 0.5


# --- Pulseaudio config ------------------------------------------------------
#
# Source: audio input connected to speaker out of the TRX
# Sink:   audio output connected to mic/line in of TRX
#
# Usually, on a raspy the correct pulseaudio source and sink can be guessed
# automatically. However, if the guessing does not work (e. g. due to
# unexpected names), the correct sink and/or source can be given here as a
# string. To get a list of available sinks or sources run the commands
#
#   pactl list sources short | grep -v monitor | cut -f 2
#   pactl list sinks short | cut -f 2
#
# The RasPi's built-in sink has the term 'bcm...' (meaning Broadcom) in its
# name.

trx_source = None
trx_sink = None

# volume in %
trx_source_volume = 100
trx_sink_volume = 100


# --- software parameters ----------------------------------------------------

# 1750 Hz detection
#
# Trigger level in dB; level of the tone above background ("noise")
# Min. duration of signal tone in s
trigger_level_dB = 20
trigger_time = 0.5

# Timeout in s for sending
#
# If no new PTT-ON signal is received within this time, sending is stopped.
# This timer is used to catch missed PTT-OFF signals.
timeout_send = 30

# Timeout for protecting a 1750 Hz call from internal traffic
#
# After a 1750 Hz call, the request for internal traffic LISTEN-INT-REQ is
# rejected and the answering of the external call encuraged.
timout_call_protect = 10


# --- import network config --------------------------------------------------
net_cfg = importlib.import_module("config_net")


# ----------------------------------------------------------------------------
# Server Class
# ----------------------------------------------------------------------------

class TRX_Server:
    """Server running the TRX in the OVF remote operation network.

    State Diagram
    -------------

    The TRX server has two internal states. The change of change is initiated
    by events from the clients (noticed via incoming tcp traffic). The send
    state may be terminated by a timout, too. 

    {{{ figure

    recv                                    send
    ==========================================================================
    PTT off                                 PTT on
    timer_stop()                            timer_send()
    @1750 Hz call: timer_call_protect()

              | tcp_listen
              '---------------------------------------------------------->
               @tcp_pressing_clients >= 0 (PTT-ON)

                                                              tcp_listen |
              <----------------------------------------------------------'
                                      @tcp_pressing_clients = 0 (PTT-OFF)

                                                                       
                                                              timer_send |
              <----------------------------------------------------------'
                                              @timeout_send

    }}}

    Internal Traffic Management
    ---------------------------

    Clinets can request internal traffic ("internal phone") by LISTEN-INT-REQ.
    The requesting client switches all other clients, mediated by the server,
    to LISTEN-INT. However, the server may reject the request with
    LISTEN-INT-REJ if an external QSO is likely to be ongoing. This is the
    case either if the server is in send state or if a 1750 Hz call is less
    than timout_call_protect ago.

    List of Threads
    ---------------

    There are (up to) four threads running. The timer thread runs only when a
    timer is needed. The udp_listener thread registers new clients in the list
    of known_clients.

    {{{ table

    Thread          incoming        outgoing            list            
                    traffic         traffic             management      
    ================+===============+===================+===================
    tcp_listen      PTT-ON                              ptt_pressing_clients
                    PTT-OFF                             known_clients
                    LISTEN-INT-REQ  LISTEN_INT
                                    LISTEN_INT_REJ
    ----------------+---------------+-------------------+-------------------
    udp_listen      OVF-REGISTER    OVF-REGISTER-OK     known_clients   
    ----------------+---------------+-------------------+-------------------
    audio_fft                       LISTEN              ptt_pressing_clients
    -> callback_1750                                    
    ----------------+---------------+-------------------+-------------------
    timer                                               ptt_pressing_clients
    ----------------+---------------+-------------------+-------------------
    }}}

    Client-Server-Connection
    ------------------------

    The clients looks for the server on udp with an OVF-REGISTER broadcast.
    The server answers over tcp with OVF-REGISTER-OK. The client checks the
    presence of the server periodically via tcp with a PING. If a server gets
    a PING, the sender is also added to known_clients -- obviously the server
    was unreachable for a short time and the client didn't notice.
    """


    def __init__(self):
        # private variables
        self.state_send = Event()
        self.state_lock = RLock()
        # public variables; server interface for threads
        self.ptt_pressing_clients = thread_set.Thread_Set()
        self.known_clients = thread_set.Thread_Set()

        # modules
        self.ptt = physical_ptt.get_ptt(ptt_mode, ptt_port, ptt_hold_on)
        self.pa_rtp = pa_rtp.PA_RTP(net_cfg.rtp_mcast_ip,
                                    trx_source,
                                    trx_sink,
                                    trx_source_volume,
                                    trx_sink_volume)
        self.pa_rtp.rx_mute(False)
        self.pa_rtp.tx_mute(False)

        # Timer
        self.timer = None
        self.timer_lock = RLock()
        self.call_protect = Event()

        # Threads for network communication
        self.th_tcp_listen = Thread(target=self.tcp_listen, daemon=True)
        self.th_udp_listen = Thread(target=self.udp_listen, daemon=True)

        # Audio thread for call tone detection (1750 Hz)
        self.th_audio_fft = th_audio_fft.Audio_FFT(self.callback_1750,
                                                trigger_level_dB,
                                                trigger_time)

    def done(self):
        # call done() for all instance attribues implementing it
        self.pa_rtp.done()
        self.ptt.done()
        self.th_audio_fft.done()

    def run(self):
        # start threads
        self.th_tcp_listen.start()
        self.th_udp_listen.start()
        self.th_audio_fft.start()
        # Wait for ending streams and terminate
        # This should never happen.
        while True:
            if not self.th_audio_fft.is_alive(): 
                raise RuntimeError('audio_fft not active')
            if not self.th_tcp_listen.is_alive(): 
                raise RuntimeError('th_tcp_listen not alive')
            if not self.th_udp_listen.is_alive(): 
                raise RuntimeError('th_udp_listen not alive')
            time.sleep(0.2)

    # ------------------------------------------------------------------------
    # state handling, thread safe
    # ------------------------------------------------------------------------

    def set_state_recv(self, call_1750=False):
        if verbose: print('set_state_recv()')
        with self.state_lock:
            self.state_send.clear()
            self.ptt.off()
            self.timer_stop()
            if call_1750: self.timer_call_protect()

    def set_state_send(self):
        if verbose: print('set_state_send()')
        with self.state_lock:
            self.state_send.set()
            self.ptt.on()
            self.timer_send()

    # ------------------------------------------------------------------------
    # Timer
    # ------------------------------------------------------------------------
    # Actually two timers are involved: One for timer_send and one for
    # timer_call_protect. Combined start and stop functions are used to assure
    # that only one of them is running at once. call_protect indicates whether
    # timer_call_protect is running.

    def timer_stop(self):
        if verbose: print('timer_stop()')
        with self.timer_lock:
            if self.timer: self.timer.cancel()
            self.call_protect.clear()

    def timer_send(self):
        if verbose: print('timer_send() at ', datetime.datetime.now())
        with self.timer_lock:
            if self.timer: self.timer.cancel()
            self.timer = Timer(timeout_send, self._timer_send_expired)
            self.timer.start()
            self.call_protect.clear()

    def timer_call_protect(self):
        if verbose: print('timer_call_protect()')
        with self.timer_lock:
            if self.timer: self.timer.cancel()
            self.timer = Timer(timout_call_protect, 
                               self._timer_call_protect_expired)
            self.timer.start()
            self.call_protect.set()

    def _timer_send_expired(self):
        with self.timer_lock:
            self.set_state_recv()

    def _timer_call_protect_expired(self):
        with self.timer_lock:
            if verbose: print('_timer_call_protect_expired() at ', 
                                                    datetime.datetime.now())
            self.call_protect.clear()

    # ------------------------------------------------------------------------
    # Thread: TCP-Listen 
    # Listen for PTT-ON, PTT-OFF, LISTEN-INT-REQ; manage ptt_pressing_clients
    # ------------------------------------------------------------------------
    #
    # Whenever a client is sending a PTT-ON message, it is noted in a list. As
    # soon as a PTT-OFF message from the same client is received, the
    # corresponging entry is deleted. If there is at least one clinet pressing
    # PTT, state 'send' is entered. If no clinet is pressing PTT, state 'recv'
    # is entered.
    #
    # LISTEN-INT-REQ are answerde either with LISTEN-INT-REJ, of all clients
    # are informed via LISTEN-INT.
    # ------------------------------------------------------------------------

    def tcp_listen(self):
        # open TCP socket for listening
        # multiple connects are handled sequentially
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((net_cfg.tcp_server_bind_ip, net_cfg.tcp_server_port))
        s.listen(5) 

        # forever: handle incoming connections sequentially
        while True:
            client, address = s.accept()
            client_ip = address[0]      # address = ('ip-address', port<int>)
            with client:
                # Read all data until connection is closed.
                message = b''
                while True:
                    # read message, might come in multiple chunks
                    chunk = client.recv(1024)
                    if not chunk: break
                    message += chunk
                    # limit message to max length to avoid DoS attac
                    if len(message) > 255:
                        client.close()
                        break
                if verbose: print(f'TCP_listener: read {message}')
                # Evaluate message
                if message == b'PING':
                    # client knew us from last run
                    self.known_clients.add(client_ip) 
                elif message == b'PTT-ON':
                    self.ptt_pressing_clients.add(client_ip)
                    self.set_state_send()
                elif message == b'PTT-OFF':
                    self.ptt_pressing_clients.discard(client_ip)
                    if len(self.ptt_pressing_clients) == 0:
                        self.set_state_recv()
                elif message == b'LISTEN-INT-REQ':
                    if self.state_send.is_set() or self.call_protect.is_set():
                        self.send_LISTEN_INT_REJ(client_ip)
                    else:
                        self.send_all_LISTEN_INT()

    # ------------------------------------------------------------------------
    # Thread: UDP-Listen 
    # Listen for new clients and register them.
    # ------------------------------------------------------------------------
    #
    # New clients ask for registration by sending an UDP broadcast
    # OVF-REGISTER. They are put to the server list of known_clients and get
    # over TCP an acknowledgement OVF-REGISTER-OK.
    # ------------------------------------------------------------------------

    def udp_listen(self):
        # open UDP socket for listening
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(('', net_cfg.udp_server_port))

        # forever: handle incoming connections
        while True:
            msg, addr = s.recvfrom(1024)
            if verbose: print(f'UDP_listener: read {msg} from {str(addr)}')
            if msg == b'OVF-REGISTER':
                client_ip = addr[0]      # address = ('ip-address', port<int>)
                self.known_clients.add(client_ip)
                self.send_OVF_REGISTER_OK(client_ip)

    # ------------------------------------------------------------------------
    # networking: sent tcp
    # ------------------------------------------------------------------------

    def send_all_LISTEN(self):
        """Send LISTEN to all known clients"""
        known_clients = self.known_clients.copy()
        for ip in known_clients:
            self._send_TCP(b'LISTEN', ip)

    def send_all_LISTEN_INT(self):
        """Send LISTEN-INT to all known clients"""
        known_clients = self.known_clients.copy()
        for ip in known_clients:
            self._send_TCP(b'LISTEN-INT', ip)

    def send_LISTEN_INT_REJ(self, ip):
        self._send_TCP(b'LISTEN-INT-REJ', ip)

    def send_OVF_REGISTER_OK(self, ip):
        self._send_TCP(b'OVF-REGISTER-OK', ip)

    def _send_TCP(self, msg, ip):
        if verbose: print(f'_send_TCP(): {msg} to {ip}')
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.settimeout(1.0)
                s.connect((ip, net_cfg.tcp_client_port))
                s.sendall(msg)
        except (ConnectionRefusedError, TimeoutError, OSError) as err:
            if verbose:
                print('_send_TCP(): connection failed, client dropped')
                print(f'            {err}, {type(err)}')
            self.known_clients.discard(ip)

    # ------------------------------------------------------------------------
    # call tone handling
    # ------------------------------------------------------------------------

    def callback_1750(self):
        self.set_state_recv(call_1750=True)
        self.send_all_LISTEN()
        self.ptt_pressing_clients.clear()

# ----------------------------------------------------------------------------
# Main Program
# ----------------------------------------------------------------------------

try:
    my_server = TRX_Server()
    my_server.run()
except KeyboardInterrupt:
    pass
finally:
    my_server.done()


# vim: set et fdm=marker:
