#!/usr/bin/python3
import pyaudio
import numpy as np
import math

verbose = False

# ----------------------------------------------------------------------------
# Thread: Call tone detection
# - Call tone (1750 Hz) detection
# - Trigger of listening clients
# ----------------------------------------------------------------------------
#
# The call tone detection is done in the callback funktion of a PyAudio audio
# stream. The call tone is detected via FFT of the audio signal.
# ----------------------------------------------------------------------------

class Audio_FFT():

    def __init__(self, callback_1750, trigger_level_dB, trigger_time):
        """Listen on incoming audio, detecting 1750 Hz call tone.

        callback_1750()
            function to call if call tone is recognised

        trigger_level_dB : float
            trigger threshold in dB, level above noise

        trigger_time : float
            min length of call tone in seconds
        """
        self.callback_1750 = callback_1750
        self.trigger_level_dB = trigger_level_dB

        self.pyaudio = None
        self.audio_stream = None
        self.trigger_counter = 0
        self.triggered = False

        # Calculation of reasonable frame length
        # {{{ Doku
        #
        # call tone                 f   = 1750 Hz
        # sample rate               f_s = 8000 Hz
        # number of samples         N
        # frame duration            T   = N / f_s
        # freq. channel step        f_c = 1 / T
        # call tone freq. channel   i   = f / f_c
        #
        # Relation of freq. channel i and frame length N:
        #
        #    i         f       T * f       f
        #   --- = --------- = -------- = -----
        #    N     f_c * N     T * f_s    f_s
        #
        # Meaningful values for actual frequencies:
        #
        #    i      f      1750 Hz      7     chosen      35
        #   --- = ----- = --------- = ----       =      -----
        #    N     f_s     8000 Hz     23                160
        #
        # Arising values:
        #   N = 160
        #   T = N / f_s = 20 ms
        #   i = 35
        #
        # }}}
        self.sampling_rate_f_s = 8000
        self.frame_len_N = 160
        self.signal_channel_i = 35

        # Calculation of 0 dB
        # {{{ Doku
        #
        # max. peak-peak amplitude in int16:                Upp = 2**16
        # => max. ampltude                                  U   = 2**15
        # decrease by 2 due to hanning-window               U   = 2**14
        # decrease by 2 due to split on pos/neg frequencies U   = 2**13 
        #
        # FFT scales amplitudes by frame length.
        #
        # => maximum amplitude in FFT: 2**13 * frame_len_N
        # }}}
        self.offset_0_dB = 20 * math.log10(2**13 * self.frame_len_N)

        # trigger time in frames
        self.trigger_frames = int(trigger_time 
                                  / self.frame_len_N 
                                  * self.sampling_rate_f_s)

        # Windowing funtion for FFT
        self.hanning = np.hanning(self.frame_len_N)

        # mask for non-tone channels ("noise")
        left_edge = self.signal_channel_i - 2
        right_edge = self.signal_channel_i + 2
        index_i = np.arange(self.frame_len_N // 2 + 1)
        self.noise_mask = np.logical_or(index_i < left_edge,
                                        index_i > right_edge)

    def start(self):
        """Create and start audio thread"""
        self.pyaudio = pyaudio.PyAudio()
        self.audio_stream = self.pyaudio.open(
                                rate=self.sampling_rate_f_s,
                                channels=1,
                                format=pyaudio.paInt16,
                                input=True,
                                frames_per_buffer=self.frame_len_N,
                                start=True,
                                stream_callback=self.callback)


    def done(self):
        """Stop and close audio stream"""
        if self.audio_stream:
            if self.audio_stream.is_active(): 
                self.audio_stream.stop_stream()
            self.audio_stream.close()
        if self.pyaudio: self.pyaudio.terminate()

    def is_alive(self):
        return self.audio_stream.is_active()

    def callback(self, in_data, frame_count, time_info, status):
        # if frame count does not match, all fft logic fails -> discard frame
        if frame_count == self.frame_len_N:
            # calculate power spectrum in dB
            a = np.frombuffer(in_data, dtype=np.int16) * self.hanning
            power_spec = np.abs(np.fft.rfft(a))
            dB = 20 * np.log10(power_spec) - self.offset_0_dB

            # estimate noise level
            noise_level = np.mean(dB[self.noise_mask])

            # check signal tone level
            if dB[self.signal_channel_i] > noise_level + self.trigger_level_dB:
                self.trigger_counter += 1
                if self.trigger_counter > self.trigger_frames:
                    if verbose: print('audio_fft: trigger set')
                    self.triggered = True
            else:
                self.trigger_counter = 0
                if self.triggered:
                    if verbose: print('audio_fft: triggered')
                    self.triggered = False
                    self.callback_1750()
        return (None, pyaudio.paContinue)


# vim: set et fdm=marker:
